import json
from multiprocessing.dummy import shutdown
import os
import shutil
from typing import Union
import requests
import random

from kachery_cloud.mutable_local import get_mutable_local, set_mutable_local
from .get_kachery_cloud_dir import get_kachery_cloud_dir
from ._kacherycloud_request import _kacherycloud_request
from .store_file_local import _compute_file_hash
from ._fs_operations import _makedirs, _chmod_file
from ._access_group_encrypt import _access_group_decrypt
from .encrypt_uri import decrypt_uri


def load_file(uri: str, *, verbose: bool=False, local_only: bool=False, dest: Union[None, str]=None) -> Union[str, None]:
    if uri.startswith('sha1-enc://'):
        # While it would be nice to send the sha1-enc hash directly to the API findFile,
        # we need to do it this way so that we can check locally... because files are stored
        # locally by the sha1 hash. Unfortunately, this means two http calls.
        uri_decrypted = decrypt_uri(uri)
        return load_file(uri_decrypted, verbose=verbose, local_only=local_only, dest=dest)

    if local_only:
        return load_file_local(uri, dest=dest)
    if uri.startswith('/'):
        if os.path.exists(uri):
            if dest is not None:
                shutil.copyfile(uri, dest)
                return dest
            return uri
        else:
            return None
    if uri.startswith('sha1://'):
        x = load_file_local(uri, dest=dest)
        if x is not None:
            return x
        sha1 = uri.split('?')[0].split('/')[2]
        return _load_sha1_file_from_cloud(sha1, verbose=verbose, dest=dest)
        
    assert uri.startswith('ipfs://'), f'Invalid or unsupported URI: {uri}'
    a = uri.split('?')[0].split('/')
    assert len(a) >= 3, f'Invalid or unsupported URI: {uri}'
    cid = a[2]

    kachery_cloud_dir = get_kachery_cloud_dir()
    e = cid[-6:]
    parent_dir = f'{kachery_cloud_dir}/ipfs/{e[0]}{e[1]}/{e[2]}{e[3]}/{e[4]}{e[5]}'
    filename = f'{parent_dir}/{cid}'
    if os.path.exists(filename):
        if dest is not None:
            shutil.copyfile(filename, dest)
            return dest
        return filename

    payload = {
        'type': 'findIpfsFile',
        'cid': cid
    }
    response= _kacherycloud_request(payload)
    found = response['found']
    if found:
        url = response['url']
    else:
        return None
        # url = f'https://{cid}.ipfs.dweb.link'
        # url = f'https://cloudflare-ipfs.com/ipfs/{cid}'
        # url = f'https://ipfs.filebase.io/ipfs/{cid}'

    if verbose:
        print(f'Loading file from kachery cloud: {uri}')    
    if not os.path.exists(parent_dir):
        _makedirs(parent_dir)
    tmp_filename = f'{filename}.tmp.{_random_string(8)}'
    with requests.get(url, stream=True) as r:
        r.raise_for_status()
        with open(tmp_filename, 'wb') as f:
            for chunk in r.iter_content(chunk_size=8192):
                f.write(chunk)
    try:
        os.rename(tmp_filename, filename)
        _chmod_file(filename)
    except:
        if not os.path.exists(filename): # maybe some other process beat us to it
            raise Exception(f'Unexpected problem moving file {tmp_filename}')
    if dest is not None:
        shutil.copyfile(filename, dest)
        return dest
    return filename

def _load_sha1_file_from_cloud(sha1: str, *, verbose: bool, dest: Union[None, str]=None) -> Union[str, None]:
    payload = {
        'type': 'findFile',
        'hashAlg': 'sha1',
        'hash': sha1
    }
    response= _kacherycloud_request(payload)
    found = response['found']
    uri = f'sha1://{sha1}'
    if found:
        url = response['url']
    else:
        return None

    kachery_cloud_dir = get_kachery_cloud_dir()
    e = sha1
    parent_dir = f'{kachery_cloud_dir}/sha1/{e[0]}{e[1]}/{e[2]}{e[3]}/{e[4]}{e[5]}'
    filename = f'{parent_dir}/{sha1}'
    if verbose:
        print(f'Loading file from kachery cloud: {uri}') 
    if not os.path.exists(parent_dir):
        _makedirs(parent_dir)
    tmp_filename = f'{filename}.tmp.{_random_string(8)}'
    with requests.get(url, stream=True) as r:
        r.raise_for_status()
        with open(tmp_filename, 'wb') as f:
            for chunk in r.iter_content(chunk_size=8192):
                f.write(chunk)
    try:
        os.rename(tmp_filename, filename)
        _chmod_file(filename)
    except:
        if not os.path.exists(filename): # maybe some other process beat us to it
            raise Exception(f'Unexpected problem moving file {tmp_filename}')
    if dest is not None:
        shutil.copyfile(filename, dest)
        return dest
    return filename

def load_file_local(uri: str, *, dest: Union[None, str]=None) -> Union[str, None]:
    query = _get_query_from_uri(uri)
    assert uri.startswith('sha1://'), f'Invalid local URI: {uri}'
    a = uri.split('?')[0].split('/')
    assert len(a) >= 3, f'Invalid or unsupported URI: {uri}'
    sha1 = a[2]

    kachery_cloud_dir = get_kachery_cloud_dir()

    s = sha1
    parent_dir = f'{kachery_cloud_dir}/sha1/{s[0]}{s[1]}/{s[2]}{s[3]}/{s[4]}{s[5]}'
    filename = f'{parent_dir}/{sha1}'
    if os.path.exists(filename):
        if dest is not None:
            shutil.copyfile(filename, dest)
            return dest
        return filename
    
    if 'location' in query:
        location = query['location']
        if os.path.isabs(location) and os.path.exists(location):
            sha1_2 = _compute_file_hash(location, 'sha1')
            if sha1_2 == sha1:
                if dest is not None:
                    shutil.copyfile(location, dest)
                    return dest
                return location
    
    # check for linked file
    a_txt = get_mutable_local(f'linked_files/sha1/{sha1}')
    if a_txt is not None:
        a = json.loads(a_txt)
        path0 = a['path']
        size0 = a['size']
        mtime0 = a['mtime']
        if os.path.exists(path0):
            if os.path.getsize(path0) == size0 and os.stat(path0).st_mtime == mtime0:
                if dest is not None:
                    shutil.copyfile(path0, dest)
                    return dest
                return path0
            sha1_0 = _compute_file_hash(path0, algorithm='sha1')
            if sha1_0 == sha1:
                set_mutable_local(f'linked_files/sha1/{sha1}', json.dumps({
                    'path': path0,
                    'size': os.path.getsize(path0),
                    'mtime': os.stat(path0).st_mtime,
                    'sha1': sha1
                }))
                if dest is not None:
                    shutil.copyfile(path0, dest)
                    return dest
                return path0
            else:
                print(f'Warning: sha1 of linked file has changed: {path0} {uri}')
    
    return None

def _get_query_from_uri(uri: str):
    a = uri.split('?')
    ret = {}
    if len(a) < 2: return ret
    b = a[1].split('&')
    for c in b:
        d = c.split('=')
        if len(d) == 2:
            ret[d[0]] = d[1]
    return ret

def _random_string(num_chars: int) -> str:
    chars = 'abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789'
    return ''.join(random.choice(chars) for _ in range(num_chars))


from typing import Union

from .get_project_id import get_project_id
from ._kacherycloud_request import _kacherycloud_request


def set_mutable(key: str, value: str, *, project_id: Union[str, None]=None):
    if len(value) > 100000:
        raise Exception('Value too large for mutable')
    payload = {
        'type': 'setMutable',
        'mutableKey': key,
        'value': value
    }
    if project_id is not None:
        payload['projectId'] = project_id
    else:
        payload['projectId'] = get_project_id()
    response = _kacherycloud_request(payload)

def get_mutable(key: str, *, project_id: Union[str, None]=None):
    payload = {
        'type': 'getMutable',
        'mutableKey': key
    }
    if project_id is not None:
        payload['projectId'] = project_id
    else:
        payload['projectId'] = get_project_id()
    response = _kacherycloud_request(payload)
    if not response['found']:
        return None
    return response['value']


from os import stat
from typing import Any, Callable, Union

from ..get_project_id import get_project_id
from .._kacherycloud_request import _kacherycloud_request
from .upload_task_result import upload_task_result
from .._serialize import _serialize

def _run_task(*, task_type: str, task_name: str, task_job_id: str, task_function: Callable, task_input: Any, project_id: Union[str, None]):
    _set_task_status(task_name=task_name, task_job_id=task_job_id, status='started', project_id=project_id)
    try:
        result = task_function(**task_input)
        result = _serialize(result)
    except Exception as err:
        _set_task_status(task_name=task_name, task_job_id=task_job_id, status='error', error=str(err), project_id=project_id)
        return False
    if task_type == 'calculation':
        upload_task_result(task_name=task_name, task_job_id=task_job_id, task_result=result, project_id=project_id)
    _set_task_status(task_name=task_name, task_job_id=task_job_id, status='finished', project_id=project_id)
    return True

def _set_task_status(*, task_name: str, task_job_id: str, status: str, error: Union[str, None]=None, project_id: Union[str, None]):
    print(f'Task {status}: {task_name}')
    payload = {
        'type': 'publishToPubsubChannel',
        'channelName': 'provideTasks',
        'message': {
            'type': 'setTaskStatus',
            'taskName': task_name,
            'taskJobId': task_job_id,
            'status': status
        }
    }
    if error is not None:
        payload['message']['errorMessage'] = error
    if project_id is not None:
        payload['projectId'] = project_id
    else:
        payload['projectId'] = get_project_id()
    _kacherycloud_request(payload)


import os

def get_kachery_cloud_dir():
    from pathlib import Path
    homedir = str(Path.home())
    hsd = os.getenv('KACHERY_CLOUD_DIR', f'{homedir}/.kachery-cloud')
    if not os.path.exists(hsd):
        os.makedirs(hsd)
    return hsd


from ._json_stringify_deterministic import _json_stringify_deterministic
from ._sha1_of_string import _sha1_of_string


def sha1_of_dict(x: dict):
    a = _json_stringify_deterministic(x)
    return _sha1_of_string(a)


import os
import json
import hashlib
import shutil
import random
from urllib.parse import quote
from typing import Union
from .get_kachery_cloud_dir import get_kachery_cloud_dir
from .mutable_local import get_mutable_local, set_mutable_local
from ._sha1_of_string import _sha1_of_string
from ._fs_operations import _makedirs, _chmod_file


def store_file_local(filename: str, *, label: Union[str, None]=None, reference: Union[bool, None]=None):
    if not os.path.isabs(filename):
        filename = os.path.abspath(filename)
    sha1 = _compute_file_hash(filename, algorithm='sha1')
    uri = f'sha1://{sha1}'
    if label is not None:
        uri = f'{uri}?label={quote(label)}'
    if reference:
        delim = '&' if '?' in uri else '?'
        uri = f'{uri}{delim}location={filename}'
        return uri
    kachery_cloud_dir = get_kachery_cloud_dir()
    kachery_storage_parent_dir = f'{kachery_cloud_dir}/sha1/{sha1[0]}{sha1[1]}/{sha1[2]}{sha1[3]}/{sha1[4]}{sha1[5]}'
    kachery_storage_file_name = f'{kachery_storage_parent_dir}/{sha1}'
    if not os.path.exists(kachery_storage_file_name):
        if not os.path.exists(kachery_storage_parent_dir):
            _makedirs(kachery_storage_parent_dir)
        tmp_filename = f'{kachery_storage_file_name}.{_random_string(10)}'
        shutil.copyfile(filename, tmp_filename)
        try:
            os.rename(tmp_filename, kachery_storage_file_name)
            _chmod_file(kachery_storage_file_name)
        except:
            # Maybe another client renamed the file
            if not os.path.exists(kachery_storage_file_name):
                raise Exception(f'Unexpected problem renaming file: {tmp_filename} {kachery_storage_file_name}')
    return uri

def _compute_file_hash(path: str, algorithm: str) -> str:
    if not os.path.isabs(path):
        path = os.path.abspath(path)
    if not os.path.exists(path):
        raise Exception(f'File does not exist: {path}')
    size0 = os.path.getsize(path)
    if size0 > 1000 * 1000:
        a = get_mutable_local(f'compute_sha1_cache/{_sha1_of_string(path)}')
        if a:
            a = json.loads(a)
            mtime = os.stat(path).st_mtime
            if a['size'] == size0 and a['mtime'] == mtime:
                return a['sha1']
    if (size0 > 1000 * 1000 * 100):
        print('Computing {} of {}'.format(algorithm, path))
    BLOCKSIZE = 65536
    hashsum = getattr(hashlib, algorithm)()
    with open(path, 'rb') as file:
        buf = file.read(BLOCKSIZE)
        while len(buf) > 0:
            hashsum.update(buf)
            buf = file.read(BLOCKSIZE)
    ret = hashsum.hexdigest()
    if size0 > 1000 * 1000:
        set_mutable_local(f'compute_sha1_cache/{_sha1_of_string(path)}', json.dumps({
            'path': path,
            'size': size0,
            'mtime': os.stat(path).st_mtime,
            'sha1': ret
        }))
    return ret

def _random_string(num_chars: int) -> str:
    chars = 'abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789'
    return ''.join(random.choice(chars) for _ in range(num_chars))


import os
from typing import Union
from .get_kachery_cloud_dir import get_kachery_cloud_dir
from ._fs_operations import _makedirs, _chmod_file


def set_mutable_local(key: str, value: str):
    assert _is_valid_key(key), f'Invalid mutable local key: {key}'
    kachery_cloud_dir = get_kachery_cloud_dir()
    file_name = f'{kachery_cloud_dir}/mutables/{_add_snails(key)}'
    parent_dir = os.path.dirname(file_name)
    if not os.path.exists(parent_dir):
        _makedirs(parent_dir)
    with open(file_name, 'w') as f:
        f.write(value)
    _chmod_file(file_name)

def _is_valid_key(key: str):
    a = key.split('/')
    for b in a:
        if b == '.' or b == '..':
            return False
        if b.startswith('@'):
            return False
    return True

def get_mutable_local(key: str, default_value: Union[None, str]=None):
    assert _is_valid_key(key), f'Invalid mutable local key: {key}'
    kachery_cloud_dir = get_kachery_cloud_dir()
    file_name = f'{kachery_cloud_dir}/mutables/{_add_snails(key)}'
    if not os.path.exists(file_name):
        return default_value
    with open(file_name, 'r') as f:
        return f.read()

def _add_snails(x: str):
    # some/path/to/key -> @some/@path/@to/key
    a = x.split('/')
    for i in range(len(a) - 1):
        a[i] = '@' + a[i]
    return '/'.join(a)


import os
import json
from urllib.parse import quote
from typing import Union
from .mutable_local import set_mutable_local
from .store_file_local import _compute_file_hash


def link_file(filename: str, *, label: Union[str, None]=None):
    if not os.path.isabs(filename):
        filename = os.path.abspath(filename)

    # this will be efficient on subsequent calls because of lookup
    sha1 = _compute_file_hash(filename, algorithm='sha1')

    size0 = os.path.getsize(filename)
    set_mutable_local(f'linked_files/sha1/{sha1}', json.dumps({
        'path': filename,
        'size': size0,
        'mtime': os.stat(filename).st_mtime,
        'sha1': sha1
    }))

    uri = f'sha1://{sha1}'
    if label is not None:
        uri = f'{uri}?label={quote(label)}'
    return uri


from typing import Union
import requests
import simplejson
import time

from ..get_project_id import get_project_id
from .._serialize import _serialize, _deserialize
from .._json_stringify_deterministic import _json_stringify_deterministic
from .._get_project_bucket_base_url import _get_project_bucket_base_url
from .._kacherycloud_request import _kacherycloud_request

def upload_task_result(*, task_name: str, task_job_id: dict, task_result: dict, project_id: Union[str, None]=None):
    task_result_serialized = _serialize(task_result)
    task_result_text = _json_stringify_deterministic(task_result_serialized)
    size = len(task_result_text)

    payload = {
        'type': 'initiateTaskResultUpload',
        'taskName': task_name,
        'taskJobId': task_job_id,
        'size': size
    }
    if project_id is not None:
        payload['projectId'] = project_id
    else:
        payload['projectId'] = get_project_id()
    response = _kacherycloud_request(payload)
    signed_upload_url = response['signedUploadUrl']
    resp_upload = requests.put(signed_upload_url, data=task_result_text)
    if resp_upload.status_code != 200:
        raise Exception(f'Error uploading task_result to bucket ({resp_upload.status_code}) {resp_upload.reason}: {resp_upload.text}')
    payload2 = {
        'type': 'finalizeTaskResultUpload',
        'taskName': task_name,
        'taskJobId': task_job_id,
        'size': size
    }
    if project_id is not None:
        payload2['projectId'] = project_id
    else:
        payload2['projectId'] = get_project_id()
    response2 = _kacherycloud_request(payload2)

def download_task_result(*, task_name: str, task_job_id: dict, project_id: str):
    project_bucket_base_url = _get_project_bucket_base_url(project_id)
    s = task_job_id
    url = f'{project_bucket_base_url}/taskResults/{task_name}/{s[0]}{s[1]}/{s[2]}{s[3]}/{s[4]}{s[5]}/{s}'
    resp = requests.get(url)
    if resp.status_code != 200:
        raise Exception(f'Error downloading task result ({resp.status_code}) {resp.reason}: {resp.text}')
    task_result = simplejson.loads(resp.content)
    task_result_deserialized = _deserialize(task_result)
    return task_result_deserialized


from setuptools import setup, find_packages

setup(
    packages=find_packages(),
    scripts=[
        'bin/kachery-cloud',
        'bin/kachery-cloud-store',
        'bin/kachery-cloud-link',
        'bin/kachery-cloud-load',
        'bin/kachery-cloud-init',
        'bin/kachery-cloud-cat',
        'bin/kachery-cloud-store-local'
    ],
    include_package_data = True,
    install_requires=[
        'requests',
        'click',
        'simplejson',
        'cryptography',
        'pubnub',
        'dask[distributed]'
    ]
)



import os
import shutil
import requests
from typing import Union
from urllib.parse import quote

from .get_kachery_cloud_dir import get_kachery_cloud_dir

from ._kacherycloud_request import _kacherycloud_request
from .get_project_id import get_project_id
from .load_file import _random_string
from ._fs_operations import _makedirs, _chmod_file
from .store_file_local import _compute_file_hash

def store_file(filename: str, *, label: Union[str, None]=None, cache_locally: bool=False):
    size = os.path.getsize(filename)
    alg = 'sha1'
    hash0 = _compute_file_hash(filename, algorithm=alg)
    uri = f'{alg}://{hash0}'
    payload = {
        'type': 'initiateFileUpload',
        'size': size,
        'hashAlg': alg,
        'hash': hash0,
        'projectId': get_project_id()
    }
    response = _kacherycloud_request(payload)
    already_exists = response['alreadyExists']
    if already_exists:
        if label is not None:
            uri = f'{uri}?label={quote(label)}'
        return uri
    signed_upload_url = response['signedUploadUrl']
    object_key = response['objectKey']
    with open(filename, 'rb') as f:
        resp_upload = requests.put(signed_upload_url, data=f)
        if resp_upload.status_code != 200:
            print(signed_upload_url)
            raise Exception(f'Error uploading file to bucket ({resp_upload.status_code}) {resp_upload.reason}: {resp_upload.text}')
    payload2 = {
        'type': 'finalizeFileUpload',
        'objectKey': object_key,
        'hashAlg': alg,
        'hash': hash0,
        'projectId': get_project_id()
    }
    response2 = _kacherycloud_request(payload2)

    if cache_locally:
        kachery_cloud_dir = get_kachery_cloud_dir()
        e = hash0
        cache_parent_dir = f'{kachery_cloud_dir}/hash0/{e[0]}{e[1]}/{e[2]}{e[3]}/{e[4]}{e[5]}'
        if not os.path.exists(cache_parent_dir):
            _makedirs(cache_parent_dir)
        cache_filename = f'{cache_parent_dir}/{hash0}'
        if not os.path.exists(cache_filename):
            tmp_filename = f'{cache_filename}.tmp.{_random_string(8)}'
            shutil.copyfile(filename, tmp_filename)
            try:
                os.rename(tmp_filename, filename)
                _chmod_file(filename)
            except:
                if not os.path.exists(cache_filename):
                    raise Exception(f'Problem renaming file: {tmp_filename} {filename}')

    if label is not None:
        uri = f'{uri}?label={quote(label)}'
    return uri